from xml.etree.ElementTree import XMLParser
__all__ = ["orthoxml_to_newick"]


class TaxonNHXMixin:
    def get_tax_nhx(self):
        tags = []
        if self.level:
            tags.append(":S={}".format(self.level))
        if self.taxid:
            tags.append(":T={}".format(self.taxid))
        return tags


class Speciation:
    type = None

    def __init__(self, parent=None):
        self.level = ""
        self.taxid = None
        self.children = []
        self.parent = parent
        if parent is not None:
            parent.add_child(self)

    def add_child(self, e):
        self.children.append(e)

    def set_level(self, level):
        self.level = level

    def set_taxid(self, taxid):
        self.taxid = taxid

    def get_newick_node_name(self):
        if not hasattr(self, 'get_tax_nhx'):
            return self.level.replace(' ', '_')
        return ""

    def as_nhx(self):
        nhx = "[&&NHX"
        t = ",".join([c.as_nhx() for c in self.children])
        if t != "":
            t = "({})".format(t)
        tags = self.get_newick_node_name()

        if self.type:
            nhx += ":Ev={}".format(self.type)
        if hasattr(self, "get_tax_nhx"):
            nhx += "".join(self.get_tax_nhx())
        nhx += "]"
        if len(nhx) > 7:
            tags += nhx
        return "{}{}".format(t, tags)


class Duplication(Speciation):
    type = "duplication"


class Leaf(Speciation):
    def __init__(self, xref, species, parent=None):
        super().__init__(parent=parent)
        self.name = xref
        self.level = species

    def get_newick_node_name(self):
        return self.name


class NHXSpeciation(Speciation, TaxonNHXMixin):
    pass

class NHXDuplication(Duplication, TaxonNHXMixin):
    pass

class NHXLeaf(Leaf, TaxonNHXMixin):
    pass


class OrthoxmlToNewick:

    def __init__(self, xref_tag="protId", encode_levels_as_nhx=True, return_gene_to_species=False):
        self.xref_tag = xref_tag
        self.gene2xref = {}
        self.trees = {}
        self.depth = 0
        self.famid = None
        self.cur_event = None
        self.cur_species = None
        self._use_nhx = encode_levels_as_nhx
        self._return_gene_to_species= return_gene_to_species

    def start(self, tag, attrib):
        if tag == "{http://orthoXML.org/2011/}species":
            self.cur_species = attrib['name']
        if tag == "{http://orthoXML.org/2011/}gene":
            self.gene2xref[attrib['id']] = (attrib[self.xref_tag], self.cur_species)
        elif tag == "{http://orthoXML.org/2011/}geneRef":
            leaf_cls = NHXLeaf if self._use_nhx else Leaf
            self.cur_event.add_child(leaf_cls(*self.gene2xref[attrib['id']]))
        elif tag == "{http://orthoXML.org/2011/}orthologGroup":
            if self.depth == 0:
                self.famid = attrib['id']
            speciation_cls = NHXSpeciation if self._use_nhx else Speciation
            self.cur_event = speciation_cls(self.cur_event)
            self.depth += 1
        elif tag == "{http://orthoXML.org/2011/}paralogGroup":
            dupl_cls = NHXDuplication if self._use_nhx else Duplication
            self.cur_event = dupl_cls(self.cur_event)
        elif tag == "{http://orthoXML.org/2011/}property":
            if attrib['name'] == "TaxRange":
                self.cur_event.set_level(attrib['value'])
            elif attrib['name'].lower() in ("taxid", "taxonid", "taxon_id", "ncbi_taxon_id"):
                self.cur_event.set_taxid(attrib['value'])

    def end(self, tag):
        if tag == "{http://orthoXML.org/2011/}paralogGroup":
            self.cur_event = self.cur_event.parent
        elif tag == "{http://orthoXML.org/2011/}orthologGroup":
            self.depth -= 1
            if self.depth == 0:
                assert(self.cur_event.parent is None)
                self.trees[self.famid] = self.cur_event.as_nhx() + ";"
            self.cur_event = self.cur_event.parent

    def close(self):
        if self._return_gene_to_species:
            gene2species = {k[0]: k[1] for k in self.gene2xref.values()}
            return self.trees, gene2species
        return self.trees


def orthoxml_to_newick(filename, xref_tag="protId", encode_levels_as_nhx=False, return_gene_to_species=False):
    """function to convert all HOGs from an orthoxml file into newick trees

    This function converts all toplevel orthologGroups into a dictionary of newick trees.
    Duplication nodes are labeled as such using the nhx tag, e.g. a paralogGroup node
    will be translated into an internal node having the nhx label [&&NHX:Ev=duplication]

    :param filename: the filename of the input orthoxml file

    :param xref_tag: the attribute of the <gene> element that should be used to get as label
                     for the leaves labels.

    :param encode_levels_as_nhx: boolean flag indicating whether or not the species information
                                 of the internal and extend nodes should be returned in NHX format
                                 with the :S=<...> and :T=<...> format. otherwise, the TaxRange
                                 value will be used as newick node label for the internal nodes.

    :param return_gene_to_species: boolean flag indicating if a mapping with the gene to species
                                   should be returned.

    :returns either a dict of {roothogid: tree} where tree is in nhx format or a tuple with the
             first element being the tree dictionary and the second being a mapping from
             {gene: species}.
    """

    target = OrthoxmlToNewick(
        xref_tag=xref_tag,
        encode_levels_as_nhx=encode_levels_as_nhx,
        return_gene_to_species=return_gene_to_species)
    parser = XMLParser(target=target)
    with open(filename, 'rb') as xml:
        for chunk in xml:
            parser.feed(chunk)
    return parser.close()



import argparse
from ._wrappers import logger
from .zoo.utils import auto_open


def extract_pw_rels(args):
    from lxml import etree
    from .zoo.hog import transform
    xml = etree.parse(args.orthoxml)
    with auto_open(args.out, 'wt') as fout:
        for p1, p2 in transform.iter_pairwise_relations(xml, rel_type=args.type, id_attribute="protId"):
            fout.write(f"{p1}\t{p2}\n")


def main():
    parser = argparse.ArgumentParser(description="FastOMA helper scripts")
    parser.add_argument('-v', default=0, action="count", help="increase verbosity")
    subparsers = parser.add_subparsers(required=True)

    parser_pw = subparsers.add_parser('pw-rel')
    parser_pw.add_argument("--type", choices=("ortholog", "paralog"), default="ortholog",
                           help="Type of relations to extract. either 'ortholog' or 'paralog'")
    parser_pw.add_argument("--out", required=True, help="Path to output file")
    parser_pw.add_argument("--orthoxml", required=True, help="Path to input orthoxml file")
    parser_pw.set_defaults(func=extract_pw_rels)

    conf = parser.parse_args()
    logger.setLevel(level=30 - 10 * min(conf.v, 2))
    logger.debug(conf)
    conf.func(conf)


if __name__ == "__main__":
    main()


import sys
from ete3 import Tree
import os
import collections

from . import _utils_roothog
from ._wrappers import logger
from . import __version__ as fastoma_version

"""

fastoma-check-input --proteomes proteome --species-tree species_tree.nwk --out-tree species_tree_checked.nwk --hogmap hogmap -vv

"""

def check_proteome_files(proteome_folder):
    proteome_files = os.listdir(proteome_folder)
    logger.info("There are %d files in the proteome folder.", len(proteome_files))

    not_fa = [i for i in proteome_files if not (i.endswith(".fa") or i.endswith(".fasta"))]
    if not_fa:
        logger.warning("We expect that only fa/fasta files are in the proteome folder. Better to remove these %s", not_fa)

    fa_fasta = [i.split(".")[-1] for i in proteome_files if i.endswith(".fa") or i.endswith(".fasta")]
    if len(fa_fasta) <= 2:
        logger.error("There are not enough proteomes in the folder ")
        logger.error("Check input failed. FastOMA halted!")
        return False

    if len(set(fa_fasta)) > 1:
        logger.warning("We expect that all fasta files are with the same format either fa or fasta, we won't include some of them %s.", proteome_files)
        return False

    species_names = [os.path.splitext(i)[0] for i in proteome_files if i.endswith(".fa") or i.endswith(".fasta")]
    return species_names


def check_proteome(species_names, prot_recs_lists, proteome_folder):
    # ids_set=set()
    isOk = True
    num_prots_all = 0
    for species_name in species_names:
        prot_recs_list = prot_recs_lists[species_name]
        prot_recs_ids = [prot_rec.id for prot_rec in prot_recs_list]
        if len(prot_recs_ids)!= len(set(prot_recs_ids)):
            logger.error("It seems that at least of the protein id in fasta file %s is repeated. Note that only part of a fasta record before space is considered.",species_name)
            logger.error("Check input failed. FastOMA halted!")
            return False
        for prot_rec_id in prot_recs_ids:
            if len(prot_rec_id)>60:   # todo 85 is the limit without considering ||s.
                logger.error("The protein ID %s is too long in species %s.fa, which should be changed. Please make sure it will be still unique. ",prot_rec_id,species_name) # (root cause issue due to truncatation by fastTree)
                logger.error("Check input failed. FastOMA halted!")
                return False


        num_prots = len(prot_recs_list) # >EP00159_Fonticula_alba_P004948_XP_009497064.1_small_nuclear_ribonucleoprotein_B_and_B'_Fonticula_alba||691883||1155004948
        num_prots_all += num_prots
        # todo , check duplicated Ids should it be done on all species ?

        if num_prots <= 2:
            logger.error("The input proteome looks empty or too few proteins or Bio.SeqIO couldn't read it, %s/%s.fa", proteome_folder, species_name)
            isOk = False
    # todo write new protoems with cleaned record ids, keep the mapping, to be used in orthoxml writing
    # use the mapping back for orhtoxml
    logger.info("There are %d proteins in total in the input proteome folder.", num_prots_all)
    return isOk


def check_hogmap_files(hogmap_folder):
    hogmap_files = os.listdir(hogmap_folder)
    logger.info("There are %d files in the hogmap folder.", len(hogmap_files))
    not_hogmap = [i for i in hogmap_files if not i.endswith(".hogmap")]
    if not_hogmap:
        logger.warning("We expect that only hogmap files are in the %s folder. Better to remove these %s",
                       hogmap_folder, not_hogmap)

    species_hogmaps = ["".join(i.split(".")[:-2]) for i in hogmap_files if i.endswith(".hogmap")]
    logger.info("There are "+str(len(species_hogmaps))+" hogmaps.")
    return species_hogmaps



def check_speciestree_internalnode(species_tree):

    # All the internal node of the input species tree should have a name
    for node in species_tree.traverse(strategy="postorder"):
        if not node.is_leaf():
            internal_node_name = []
            if (not node.name) or len(node.name) < 1:
                logger.warning("one of the internal node in species tree doesn't have a name. we'll update the species tree.")
            else:
                internal_node_name.append(node.name)

    if len(internal_node_name) != len(set(internal_node_name)):
        logger.warning("All the internal node names should be unique. One of the internal node is repeated:")
        logger.warning([item for item, count in collections.Counter(internal_node_name).items() if count > 1])
        logger.warning("We'll change the internal node names.")
    return 1


def check_speciestree_leaves(species_tree, species_names):
    leaves_name = [i.name for i in species_tree.get_leaves()]
    logger.info("The species tree has %d leaves.", len(leaves_name))
    if len(set(leaves_name)) != len(leaves_name):
        logger.error("Leaves name should be unique in the species tree. You may try ete3 prune")
        logger.error("Check input failed. FastOMA halted!")
        return False

    species_names_not_intree = [i for i in species_names if i not in leaves_name]
    if species_names_not_intree:
        logger.error("These species are not in the species tree:\n %s", "\n ".join(species_names_not_intree))
        logger.error("This is a tree template:  ((AQUAE,CHLTR)inter1,MYCGE)inter2; for proteome folder with these three files:  AQUAE.fa  CHLTR.fa  MYCGE.fa. ")
        logger.error("Check input failed. FastOMA halted!")
        return False

    leaves_tree_not_proteome = [i for i in leaves_name if i not in species_names]
    if leaves_tree_not_proteome:
        logger.warning("there are %d leaves in the species tree that there is no proteome. So we will discard them.", len(leaves_tree_not_proteome))
    return True


def check_omamer_db(omamerdb_adress=None):
    if omamerdb_adress is None:
        logger.warning("omamer_db not passed. will assume you know what you are doing.")
        return True

    if os.path.exists(omamerdb_adress):
        if os.path.getsize(omamerdb_adress) > 10000:  # 3 bytes
            omamerdb = True
            # todo we can do some checks on version omamer v2
        else:
            logger.warning("The omamer db looks very small. are you sure it is correct?"+omamerdb_adress)
            omamerdb = False
    else:
        omamerdb =False
        logger.info("OMAmer db does not exist.")

    return omamerdb


def add_internal_node_prune(species_tree, species_names, out_fname):
    # add name for the internal, if no name is provided, removintg special chars
    counter_internal = 0
    node_names = set()
    for node in species_tree.traverse(strategy="postorder"):
        if not node.is_leaf():
            node_name = node.name

            if len(node_name) < 3 or node_name in node_names:
                if not node.is_leaf():
                    node.name = "internal_" + str(counter_internal)
                    counter_internal += 1
                    node_names.add(node.name)
                    logger.debug("The internal node name was too small or repeated "+node_name+" which is changed to "+node.name)
            elif not all(e.isalnum() or e in ("_", "-") for e in node_name):  #any(not c.isalnum() for c in node_name):
                node_name_new = ''.join(e for e in node_name if e.isalnum() or e in ("_", "-"))  # removing special chars
                if node_name_new in node_names:
                    node_name_new += "_" + str(counter_internal)
                    counter_internal += 1
                logger.debug("The internal node name has special chars or repeated '%s' which is changed to '%s'",
                             node.name, node_name_new)
                node.name = node_name_new
                node_names.add(node.name)
            else:
                node_names.add(node_name)

    logger.info("The species tree has %d leaves", len(species_tree))
    species_tree.prune(species_names)  # , preserve_branch_length=True)
    logger.info("After pruning, the species tree has %d leaves", len(species_tree))

    species_tree_output = "./species_tree_checked.nwk"
    species_tree.write(format=1, format_root_node=True, outfile=out_fname)
    logger.debug("The new species tree is written %s", out_fname)

    return species_tree


def check_splice(isoform_by_gene_all):
    total_genes = 0
    total_isoforms = 0
    spliced_species_num = 0
    for sp, isoforms in isoform_by_gene_all.items():
        if len(isoforms)>1:
            spliced_species_num+=1

            for isoform in isoforms:
                if len(isoform)>1:
                    total_isoforms+=len(isoform)
                    total_genes+=1

    logger.debug("For "+str(spliced_species_num)+"  species, out of "+str(len(isoform_by_gene_all))+" , we have splice files.")
    if total_genes ==0:
        logger.debug("It seems that in all of the splice files, each line has only one item. Make sure that the splitter in each line is semicolon ; ")
        sys.exit(1)

    logger.debug("In total, for "+ str(total_genes)+" genes, we have " + str(total_isoforms)+"  splices.")

    # make sys back
    return 1


def fastoma_check_input():
    import argparse
    parser = argparse.ArgumentParser(description="checking parameters for FastOMA")
    parser.add_argument("--version", action="version", version="FastOMA v"+fastoma_version)
    parser.add_argument("--proteomes", required=True, help="Path to the folder containing the input proteomes")
    parser.add_argument("--species-tree", required=True, help="Path to the input species tree file in newick format")
    parser.add_argument("--out-tree", required=True, help="Path to output file for sanitised species tree. ")
    parser.add_argument("--splice", help="Path to the folder containing the splice information files")
    parser.add_argument("--hogmap", help="Path to the folder containing the hogmap files")
    parser.add_argument("--omamer_db", help="Path to the omamer database")
    parser.add_argument('-v', action="count", default=0, help="Increase verbosity to info/debug")
    conf = parser.parse_args() # conf_check_input

    logger.setLevel(level=30 - 10 * min(conf.v, 2))
    logger.debug("Arguments: %s", conf)


    species_names = check_proteome_files(conf.proteomes)
    if not species_names:
        logger.error("Halting FastOMA because of invalid proteome input data")
        sys.exit(1)


    try:
        species_tree = Tree(conf.species_tree, format=1)
    except:
        try:
            species_tree = Tree(conf.species_tree)
            # todo add check fro Phyloxml
        except:
            logger.error("We have problem with parsing species tree %s using ete3 Tree. Maybe there are some special chars.", conf.species_tree)
            sys.exit(1)

    check_speciestree_internalnode(species_tree)
    if not check_speciestree_leaves(species_tree, species_names):
        logger.error("Halting FastOMA because of invalid species tree")
        sys.exit(1)
    add_internal_node_prune(species_tree, species_names, conf.out_tree)

    species_names, prot_recs_lists, fasta_format_keep = _utils_roothog.parse_proteomes(conf.proteomes)
    if not check_proteome(species_names, prot_recs_lists, conf.proteomes):
        logger.error("Halting FastOMA because of invalid proteome input data")
        sys.exit(1)

    hogmap_files = conf.hogmap is not None and os.path.exists(conf.hogmap)
    species_hogmaps = []
    if hogmap_files:
        species_hogmaps = check_hogmap_files(conf.hogmap)

    hogmap_complete = False
    if set(species_hogmaps) == set(species_names):
        hogmap_complete = True

    omamerdb = check_omamer_db(conf.omamer_db)

    if not(omamerdb or hogmap_complete):
        logger.error("OMAmer db does not exist and no hogmap provided.")
        logger.error("Check input failed. FastOMA halted!")
        sys.exit(1)
    else:
        logger.info("OMAmer db or hogmap exist. It looks ok.")

    # todo  check splice file format . if the name matches with the proteome files.
    splice_files = conf.splice is not None and os.path.exists(conf.splice)
    if splice_files:
        logger.debug("splice folder exist. Let's see its inside.")
        isoform_by_gene_all = _utils_roothog.parse_isoform_file(species_names, conf.splice)
        check_splice(isoform_by_gene_all)
    else:
        logger.info("Splice folder doesn't exist and that's ok.")
    logger.info("Input check finished ! ")




import os
import time
import random
from pyparsing import ParseException
import shutil
from Bio import AlignIO, SeqIO
import dendropy
from FastOMA.zoo.wrappers import WrapperError
import logging

from .parsers import GuenomuParser
from .base_treebuilder import TreeBuilder, AlignmentInput, DataType

from ..abstract_cli import AbstractCLI
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet

from ...file_utils import TempFile, TempDir

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


class GuenomuCLI(AbstractCLI):
    @property
    def _default_exe(self):
        return "guenomu"


class Guenomu(TreeBuilder):
    """
    Species tree reconstruction 
    
    This is a wrapper for the guenomu software, that estimates species trees from (distributions of) gene family trees.
    It does rely on previously built trees for the gene families (that don't need to be orthologs-only, by the way).
    """

    def __init__(self, genetrees, species, *args, **kwargs):
        """
        :param genetrees: a list of dendropy.TreeList objects (which may have one or more trees). Must be a list, even \
        for a single dendropy.TreeList
        :param species: a list of species names (as strings), that must be found within the gene leaves
        """
        self.options = get_default_options()
        self.genetrees = genetrees  # TODO: check if is a list of TreeList
        self.species = species  # TODO: check if it's a list of strings
        self.elapsed_time = None
        self.stdout = None
        self.stderr = None
        try:
            self.cli = self._init_cli()
        except IOError as err:
            raise WrapperError('Error searching for binary: {}'.format(err))
            # End setup

    def __call__(self, *args, **kwargs):
        """
        Sets up temporary output files and calls guenomu using _call() function.
        Writes temporary input files since we are working with dendropy.TreeList
        Saves the stdout and stderr and returns
        """
        start = time.time()  # time the execution

        with TempDir() as tmpd:  # creates temporary directory and removes it when out of scope (i.e. leave this block)
            os.chdir(tmpd)
            # create one nexus file per gene family tree
            glistnames = []
            for i in range(len(self.genetrees)):
                print(i)  ## DEBUG
                glistnames.append("gene" + str(i))
                self.genetrees[i].write(path="gene" + str(i), schema="nexus")
            # create file with name of all gene family tree files
            with open("gene_tree_list", 'w') as f:
                for s in glistnames:
                    f.write(s + '\n')
            # create file with names of all species
            with open("species_names", 'w') as f:
                for s in self.species:
                    f.write(s + '\n')

            self.stdout, self.stderr = self._call(tmpd, *args, **kwargs)
            self.result = self._read_result(tmpd)  # store result

        end = time.time()
        self.elapsed_time = end - start
        return self.result["best_topology"]
        # End call

    # Any other accessory methods
    def _call(self, tempdir, *args, **kwargs):
        """
        Call underlying low level wrapper.

        Options are passed via *args and **kwargs [This only
        covers the simplest automatic case]
        """
        self.cli('{} -S species_names -G gene_tree_list -z 0'.format(self.command()), wait=True)  # importance sampling
        self.cli('{} -S species_names -G gene_tree_list -z 1'.format(self.command()), wait=True)  # analyse output
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, tmpd):
        """
        Read back the result species tree files.
        """
        ## job0.<genetreefilename>.trprobs also available, as well as job0.params.txt
        expected_outfiles = [os.path.join(tmpd, 'species.tre'), os.path.join(tmpd, 'job0.unrooted.trprobs')]
        parser = GuenomuParser()
        result = parser.to_dict(expected_outfiles)
        return result

    def _init_cli(self):
        return GuenomuCLI()


def get_default_options():
    return OptionSet([
        # -L, --lambda=<double>        baseline reconciliation hyperprior that control distances (one or several values, should be << 1)
        FloatOption('-L', 0.0001, active=True),
        # -A, --sa_samples=<int>       number of cycles/samples simulated Annealing
        IntegerOption('-A', 10, active=True),
        # -a, --sa_iter=<int>          iterations per cycle for simulated Annealing
        IntegerOption('-a', 1000, active=True),
        # -t, --temp_i=<double>        initial (inverse) temperature for each cycle of simulated annealing
        FloatOption('-t', 0.1, active=True),
        # -T, --temp_f=<double>        final (inverse) temperature at each cycle of simulated annealing
        FloatOption('-T', 5., active=True),
        # -b, --burnin=<int>           iterations for burnin stage of Bayesian posterior sampling
        IntegerOption('-b', 0, active=True),  # default is no burnin since we run only Simul Annealing
        # -B, --b_iterations=<int>     iterations for main Bayesian posterior sampling stage (zero if you only want simulated annealing)
        IntegerOption('-B', 0, active=True),  # default is no burnin since we run only Simul Annealing
        # -n, --samples=<int>          Number of samples to be stored to file and analysed later
        IntegerOption('-n', 1, active=True),  # under SimulAnneal-only, this will default to number of cycles (-A)
        # -N, --printout=<int>         how many times summary info should be output on screen
        IntegerOption('-N', 1, active=True),
        # -m, --mtm=<int>              sample size (number of trials) for each Multiple-Try Metropolis proposal
        IntegerOption('-m', 4, active=True),
        # -p, --partition=<int>        auxiliary MCMC length to avoid the Partition function calculation (zero to neglect normalization)
        IntegerOption('-p', 0, active=True),
        # -D, --distances=<string>     string with seven  1's and 0's describing distances to be used [dup, los, ils, rf, hdist, hdist_final, spr]
        StringOption('-D', '1111111', active=True),
    ])


"""
guenomu  [-h|--help] [-L|--lambda=<double>]... [-A|--sa_samples=<int>] [-a|--sa_iter=<int>] [-t|--temp_i=<double>] [-T|--temp_f=<double>] [-b|--burnin=<int>] [-B|--b_iterations=<int>] [-n|--samples=<int>] [-N|--printout=<int>] [-m|--mtm=<int>] [-p|--partition=<int>] [-z|--action=<int>] [-D|--distances=<string>] [-S|--species=<file>] [-G|--genetrees=<file>] [<file>]
  -z, --action=<int>           execution mode for program: 0='importance sampling', 1='analisis of output' etc.
  -S, --species=<file>         file with species names, one name per line (nexus-style bracketed comments are allowed)
  -G, --genetrees=<file>       file with list of gene tree file names, one gene tree file name per line
  <file>                       Name of control file with remaining parameters (those defined also here overwrite those from control file)

"""



import os
import collections
import re



__all__ = ['tail', 'fall_back_tail', 'grep']


def tail(fh, lines=20, block_size=1024):
    """Returns the last n lines from a file

    This function returns the last n lines from an file-like
    object. It does this efficiently without reading the whole
    file, but rather by loading blocks from the end of the file.

    .. note::

        If the file is opened in text mode, i.e. open('/path', 'rt'),
        python3 cannot efficiently move in the file. In this case,
        the function fall back to a slow method that goes through
        the whole file.

    Example:

    >>> with open("/etc/passwd", 'rb') as f:
    ...     last_lines = tail(f, 2)
    ...
    >>> print(last_lines)

    :param fh: file-like object to read from
    :param int lines: number of lines to be returned
    :param int block_size: size of block to be read at once.
        intended for optimisation.
    :returns: The last lines as a list of bytes/str object"""

    if lines <= 0:
        raise ValueError('invalid lines value %r' % lines)

    encoded = getattr(fh, 'encoding', False)
    if encoded:
        return fall_back_tail(fh, lines)
    CR = '\n' if encoded else b'\n'
    data = '' if encoded else b''
    fh.seek(0, os.SEEK_END)
    fsize = fh.tell()
    block = -1
    loaded_enough_data = False
    while not loaded_enough_data:
        step = (block * block_size)
        if abs(step) >= fsize:
            fh.seek(0)
            newdata = fh.read(block_size - (abs(step) - fsize))
            loaded_enough_data = True
        else:
            fh.seek(step, os.SEEK_END)
            newdata = fh.read(block_size)
        data = newdata + data
        if data.count(CR) > lines:
            break
        else:
            block -= 1
    return data.splitlines()[-lines:]


def fall_back_tail(fh, lines):
    fh.seek(0)
    data = collections.deque(fh, maxlen=lines)
    return [e.rstrip('\n') for e in data]


def grep(fh, pat):
    """Yields lines matching a pattern

    This function yields all the lines that match a given pattern.
    The pattern can be either a simple str/bytes, or a compiled
    regex expression. The newline character is not removed.

    Example:
        >>> with open('/etc/hosts', 'rb') as fh:
        ...    for line in grep(fh, b'127.0.0.1'):
        ...        print(line)
        127.0.0.1       localhost

    :param fh: file-like object
    :param pat: search pattern, either str, bytes or compiled regex
    :returns: generator yielding lines matching pattern.

    """
    if isinstance(pat, (str, bytes)):
        encoded = getattr(fh, 'encoding', False)
        if encoded and isinstance(pat, bytes):
            pat = re.compile(pat.decode())
        elif not encoded and isinstance(pat, str):
            pat = re.compile(pat.encode('utf-8'))
        else:
            pat = re.compile(pat)
    fh.seek(0)
    for line in fh:
        if pat.search(line):
            yield line



import time
from Bio import AlignIO, SeqIO
import tempfile
from six import StringIO
from ..abstract_cli import AbstractCLI
from .base_aligner import Aligner, AlignmentInput, DataType
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet




class ProGraphMSACLI(AbstractCLI):
    """
    PrographMSA low-level command line interface

    :Example:

    ::

        prograph_cli = ProGraphMSACLI()
        process = prograph_cli(cmd='mafft args...')
        stdout = prograph_cli.get_stdout()
    """

    @property
    def _default_exe(self):
        return 'ProGraphMSA'


def set_default_dna_options(aligner):
    """
    Dummy function as sensible default already provided by mafft --auto
    """
    aligner.options = get_default_options()


def set_default_protein_options(aligner):
    """
    Dummy function as sensible default already provided by mafft --auto
    """
    aligner.options = get_default_options()


class ProGraphMSA(Aligner):
    """
    Convenient wrapper for ProGraphMSA multiple sequence aligner

    The wrapper is written as a callable class.
    This can hold data (state) to do with the operation it performs, so it can keep results,
    execution times and other metadata, as well as perform the task.

    This is a basic implementation that can be extended. The important parts are
    __init__ (does the setup) and __call__ (does the work). All
    else are helper methods.

    :Example:

    ::

        callable_wrapper = Mafft(aln)
        result = callable_wrapper()
        time_taken = callable_wrapper.elapsed_time
        result_again = callable_wrapper.result
    """

    def __init__(self, input_, *args, **kwargs):
        super(ProGraphMSA, self).__init__(input_, *args, **kwargs)
        self.options = get_default_options()
        if self.datatype == DataType.DNA:
            set_default_dna_options(self)
        else:
            set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Anything to do with calling ProGraphMSA should go here.
        If any extra arguments need to be passed they can
        be specified (listed as *args and **kwargs for now).
        """
        start = time.time()  # time the execution

        if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
            with tempfile.NamedTemporaryFile(mode="wt") as fh:
                SeqIO.write(self.input, fh, 'fasta')
                fh.seek(0)
                output, error = self._call(fh.name, *args, **kwargs)

        else:
            output, error = self._call(self.input, *args, **kwargs)

        self.result = self._read_result(output)  # store result
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods
    def _call(self, filename, *args, **kwargs):
        """
        Call underlying low level ProGraphMSA wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        self.cli('{} {}'.format(self.command(), filename),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, output):
        """
        Read back the result.
        """
        fileobj = StringIO(output)
        return AlignIO.read(fileobj, 'fasta')

    def _init_cli(self, binary):
        return ProGraphMSACLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # output fasta format (instead of stockholm), better because no tree output is produced
        FlagOption('--fasta', True, active=True),

        # output all ancestral sequences
        FlagOption('--ancestral_seqs', False, active=False),

        # output sequences in input order (default: tree order)
        FlagOption('--input_order', False, active=False),

        # output all intermediate guide trees
        FlagOption('--all_trees', False, active=False),

        # use ML distances with gap
        FlagOption('--mldist_gap', False, active=False),

        # use ML distances
        FlagOption('--mldist', False, active=False),

        # use of guide tree
        StringOption('--tree', '', active=False)

    ])



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_hooks()
from .genetree import *
from .taxonomy import *



import os
import shutil
import tempfile



__all__ = ['TempFile', 'TempDir', 'ChDir', 'MkDir', 'NonDeletingTempDir']

class TempFile(object):
    """ 
    Context manager for working with a temporary file
    that automatically cleans up.

    Usage:

    with TempFile() as tmp:
        # In scope, tmp exists on the disk
        # Do some work with tmp, e.g. tmp.write('something')

    # Out of scope, tmp is deleted

    with TempFile('local_temp_space') as tmp:
        # tmp is created in the directory 'local_temp_space'
        # The specified directory must exist, or an error is thrown

    """

    def __init__(self, dir_=None):
        if dir_ is not None and not os.path.exists(dir_):
            raise IOError('Directory "{}"" does not exist'.format(dir_))
        self.dir = dir_

    def __enter__(self):
        self._fd, self._wrapped_tmp = tempfile.mkstemp(dir=self.dir)
        return os.path.abspath(self._wrapped_tmp)

    def __exit__(self, type, value, tb):
        os.close(self._fd)
        os.remove(self._wrapped_tmp)


class TempDir(object):
    """
    Context manager for working with a temporary file
    that automatically cleans up.

    Usage:

    with TempDir() as tmpd:
        # In scope, tmpd exists on the disk
        # Do some work with tmpd ...

    # Out of scope, tmpd is deleted along with all its content

    Can be nested with TempFile, e.g.

    with TempDir() as tmpd, TempFile(tmpd) as tmpf:
        # tempfile tmpf is created inside temporary directory tmpd
    # On exit, everything is deleted

    """

    def __enter__(self):
        self._wrapped_tmpdir = tempfile.mkdtemp()
        return os.path.abspath(self._wrapped_tmpdir)

    def __exit__(self, type, value, tb):
        shutil.rmtree(self._wrapped_tmpdir)


class NonDeletingTempDir(TempDir):
    def __exit__(self, tpye, value, tb):
        pass


class ChDir(object):
    """
    Context manager to switch to a working directory,
    and return to the current directory (like 'Dir.chdir do' block in Ruby)

    Usage:

    with TempDir() as dir, ChDir(dir):
        # Do some work in the working temp directory 'dir'

    # Exit 'dir'
    """

    def __init__(self, working_dir):
        if not os.path.exists(working_dir):
            raise IOError('Directory "{}"" does not exist'.format(working_dir))
        self._cdir = os.getcwd()
        self._wdir = working_dir

    def __enter__(self):
        os.chdir(self._wdir)

    def __exit__(self, type, value, tb):
        os.chdir(self._cdir)


class MkDir(ChDir):
    """
    Context manager to create and switch to a working directory,
    then return to the current directory.

    Usage:

    with TempDir() as dir, MkDir(dir):
        # Do some work in the working temp directory 'dir'

    # Exit 'dir'
    """

    def __init__(self, working_dir):
        if not os.path.exists(working_dir):
            try:
                os.makedirs(working_dir)
            except OSError as e:
                if e.errno != 17:
                    raise
                pass  # path was created by another thread / process
                # this is a race condition, but probably benign

    def __enter__(self):
        pass

    def __exit__(self, type, value, tb):
        pass




import xml.etree.ElementTree as ET
from Bio.Align import MultipleSeqAlignment
from Bio.SeqRecord import SeqRecord
from random import sample
from typing import Optional, List, Union
from ete3 import Tree, TreeNode
import random
from ._utils_subhog import MSAFilter

# todo some of these could also come under conf_infer_subhhogs
seed_random=1234 # Also in _wrappers.py
random.seed(seed_random)
fragment_detection = True  # this also need to be consistent in _infer_subhog.py
fragment_detection_msa_merge = True
subsampling_hogclass = True
hogclass_max_num_seq = 10              # 40 subsampling in msa # ver very 2
hogclass_min_cols_msa_to_filter = hogclass_max_num_seq * 50
#hogclass_tresh_ratio_gap_col = 0.05     # 0.8 for very very big


import itertools
from . import _utils_subhog
from ._wrappers import logger

# from .infer_subhogs import conf_infer_subhhogs #fastoma_infer_subhogs #


class Representative:
    def __init__(self, the_one, elements:Optional[List]=None):
        self.enabled = True
        if isinstance(the_one, Representative):
            self._seq = the_one.get_record()
            self._subelements = the_one.get_subelements()
            self._species = the_one.get_species()
        elif isinstance(the_one, SeqRecord):
            self._seq = the_one
            self._subelements = set([the_one.id])
            # TODO: this should be done more genenerally. copied from singleton_hog function
            self._species = {the_one.id.split('||')[1]}
        if elements is not None:
            for e in elements:
                self._subelements |= e.get_subelements()
                self._species |= e.get_species()

    def get_id(self):
        return self._seq.id

    def get_record(self):
        return self._seq

    def get_subelements(self):
        return self._subelements

    def get_species(self):
        return self._species

    def __repr__(self):
        return '<Representative: {} [{} elements; {} species]>'.format(self.get_id(), len(self.get_subelements()), len(self.get_species()))

class HOG:
    _hogid_iter = 10000

    def __init__(self, input_instantiate, taxnomic_range:TreeNode, rhogid:str, msa:Optional[MultipleSeqAlignment] = None,
                 representatives=Optional[List[Representative]], conf_infer_subhhogs=None):
        """HOG class constructor.

        A HOG can be initialized with either a single SeqRecord or from a list of HOGs
        (which will become its sub-hog elements). Furthermore, the taxonomic range
        (a node from the species tree) needs to be provided.
        Parameters:
        :param input_instantiate"""
        # fragment_list list of  sets , each set contains protein ID of fragments
        # the input_instantiate could be either
        #     1) orthoxml_to_newick.py protein as the biopython seq record  SeqRecord(seq=Seq('MAPSSRSPSPRT. ]
        # or  2) orthoxml_to_newick.py set of intances of class HOG   wit orthoxml_to_newick.py big msa
        # those variable starting with _ are local to the class, should not access directly  (although it is possbile)
        self._rhogid = rhogid
        self.__class__._hogid_iter += 1
        # 0070124

        self._hogid = "HOG_" + self._rhogid+ "_sub" + str(self.__class__._hogid_iter)
        self._tax_least = taxnomic_range  #  least taxnomic level
        self._tax_now = taxnomic_range    # the taxnomic level that we are considering now, checking for duplication
        self.active = True

        if isinstance(input_instantiate, SeqRecord):  # if len(sub_hogs)==1:
            only_protein = input_instantiate  # only one seq, only on child, leaf
            self._members = set([only_protein.id])
            self._representatives = [Representative(only_protein)]
            self._msa = MultipleSeqAlignment([only_protein])
            self._subhogs = []
            self._dubious_members = set()
            self._num_species_tax_speciestree = 1  # at the leaf level, there is only one species in the species tree
            # <<class 'Bio.Align.MultipleSeqAlignment'> instance (1 records of length 314) at 7f0e86c713d0>

        elif msa and all(isinstance(x, HOG) for x in input_instantiate):
            # here we want to merge few subhogs and create orthoxml_to_newick.py new HOG.   #the n

            sub_hogs = input_instantiate
            hog_members = set()
            for sub_hog in sub_hogs:
                hog_members |= sub_hog.get_members()
            self._members = hog_members  # set.union(*tup)    # this also include dubious_members
            self._subhogs = list(input_instantiate)  # full members of subhog, children
            self._num_species_tax_speciestree = len(taxnomic_range.get_leaves())
            self._representatives = [r for r in representatives]
            representatives_id = set(r.get_id() for r in representatives)
            assert len(representatives_id) == len(self._representatives), "Non-unique ids among representatives"
            assert representatives_id.issubset(self._members), "Representatives must be subset of HOG members"

            dubious_members = set()
            for sub_hog in sub_hogs:
                dubious_members |= sub_hog.get_dubious_members()  # union
            self._dubious_members = dubious_members
            assert self._dubious_members.isdisjoint(representatives_id), \
                   'Dubious members has non-empty intersection with representatives:' + str(self._dubious_members.intersection(representatives_id))

            records = [record for record in msa if (record.id in representatives_id)]

            if len(records[0]) > hogclass_min_cols_msa_to_filter and conf_infer_subhhogs is not None:
                filter = MSAFilter(levelprocessor=None, conf=conf_infer_subhhogs)
                records = filter.msa_filter_col(records)
                # the challange is that one of the sequences might be complete gap
            self._msa = MultipleSeqAlignment(records)
            # without replacement sampling ,  # self._children = sub_hogs # as legacy  ?
        else:
            logger.error("Error 142769,  check the input format to instantiate HOG class")
            assert False

    def __repr__(self):
        return f"<HOG:{self.hogid},size={len(self._members)},taxLeast={self._tax_least.name},taxNow={self._tax_now.name}>"

    @property
    def hogid(self):
        return self._hogid

    @property
    def taxname(self):
        return self._tax_now.name

    @property
    def taxlevel(self):
        return self._tax_now

    @property
    def rhogid(self):
        return self._rhogid

    def get_members(self):
        return set(self._members)

    def get_dubious_members(self):
        return self._dubious_members

    def get_representatives(self):
        return self._representatives

    def get_msa(self):
        return self._msa

    def find_representative(self, seq_id):
        """Find the representative of the given sequence using its ID."""
        for r in self._representatives:
            if r.get_id() == seq_id:
                return r
        for r in self._representatives:
            if seq_id in r.get_members():
                return r
        raise KeyError("No representative found for " + seq_id)

    def remove_prot_from_hog(self, prot_to_remove):
        prot_members_hog_old = self._members
        assert prot_members_hog_old
        prot_members_hog_edited = prot_members_hog_old - set([prot_to_remove])
        self._members = prot_members_hog_edited # self._members.remove(prot_to_remove)    # discard
        msa_old = self._msa     #  we may want to edit the msa of children level to be consistent
        msa_edited = MultipleSeqAlignment([i for i in msa_old if i.id != prot_to_remove])
        self._msa = msa_edited
        if len(prot_members_hog_edited) == 0:  # hog should be removed, no members is left
            return 0
        return 1

    def insert_dubious_prots(self, fragment_host, fragments_list_nothost):

        self._dubious_members |= set(fragments_list_nothost)
        self._dubious_members |= {fragment_host}
        if fragment_host not in self._members:
            logger.error("Error 1252769, fragment_host not in hog"+str(fragment_host))
        self._members |= set(fragments_list_nothost)

        msa_old = self._msa
        msa_edited = MultipleSeqAlignment([i for i in msa_old if i.id not in self._dubious_members])
        self._msa = msa_edited
        # species_name = fragment_host.split("||")[1]   # if self._tax_now == species_name:
        return 1

    def __contains__(self, item):
        if isinstance(item, Representative):
            return item.get_id() in self._members
        elif isinstance(item, HOG):
            return item in self._subhogs
        else:
            return item in self._members

    def get_subhog_path(self, seq_id, max_depth=-1):
        if seq_id not in self._members:
            raise KeyError(f"{seq_id} is not part of this HOG")
        path = []
        h = self
        while max_depth != 0 and len(h._subhogs) > 0:
            for s in h._subhogs:
                if seq_id in s:
                    path.append(s)
                    h = s
                    break
            max_depth -= 1
        return path

    def merge_prots_name_hog(self, fragment_name_host, merged_fragment_name):
        prot_members_hog = list(self._members)
        assert fragment_name_host in prot_members_hog, str(fragment_name_host)+str("  prot_members_hog:")+str(prot_members_hog)
        fragment_host_idx = prot_members_hog.index(fragment_name_host)
        # merged_fragment_name = fragment_name_host + "_|_" + fragment_name_remove1 + ... # there might be more than one fragment
        prot_members_hog[fragment_host_idx] = merged_fragment_name
        # 'BUPERY_R03529||BUPERY||1105002086_|_BUPERY_R10933||BUPERY||1105008975']
        self._members = set(prot_members_hog)

        return 1

    def merge_prots_msa(self, merged_fragment_name, merged_msa_new):   # merged_fragment_name 'BUPERY_R03529||BUPERY||1105002086_|_BUPERY_R10933||BUPERY||1105008975']
        prot_members_hog = list(self._members)
        assert merged_fragment_name in prot_members_hog  # merged_fragment_name in self._members   -> True . already the name of host prot is changed using merge_prots_name_hierarchy_toleaves
        msa_new = []
        for seq_record in merged_msa_new:
            if seq_record.id in prot_members_hog:
                msa_new.append(seq_record)
            # else:  # some of the sequences in the msa is for other subhogs, so we only care about those in this subhog.
        # msa_old = self._msa
        # host_fragment_name = merged_fragment_name.split("_|_")[0]
        # merged_msa_new_ids = [i.id for i in merged_msa_new]         # for seq_record in msa_old:
        self._msa = MultipleSeqAlignment(msa_new)

        return 1


    # def merge_prots_msa(self, fragment_name_host, fragment_name_remove, merged_sequence, merged_msa_new):
    #     merged_fragment_name = fragment_name_host + "_|_" + fragment_name_remove
    #     prot_members_hog = list(self._members)
    #     assert merged_fragment_name in prot_members_hog
    #     # merged_fragment_name in self._members   -> True . already the name of host prot is changed using merge_prots_name_hierarchy_toleaves
    #     msa_new = []
    #     for prot_name in self._members:
    #         if seq_rec.id == fragment_name_host:
    #             seq_rec_edited = SeqRecord(Seq(merged_sequence), id=merged_fragment_name, name=merged_fragment_name)
    #         msa_new.append(seq_rec_edited)
    #     self._msa = MultipleSeqAlignment(msa_new)
    #     return 1

    def to_orthoxml(self):
        if len(self._subhogs) == 0:
            list_member = list(self._members)
            if len(list_member) == 1:
                list_member_first = list(self._members)[0]  # ['tr|A0A3Q2UIK0|A0A3Q2UIK0_CHICK||CHICK_||1053007703']
                if fragment_detection and fragment_detection_msa_merge and "_|_" in list_member_first:
                    paralog_element = ET.Element('paralogGroup')
                    #property_element = ET.SubElement(paralog_element, "property", attrib={"name": "TaxRange", "value": str(self._tax_now)})
                    property_element = ET.SubElement(paralog_element, "property", attrib={"name": "Type", "value": "DubiousMergedfragment"})
                    #todo is it happening in output?
                    # property_element = ET.SubElement(paralog_element, "property", attrib={"TaxRange": str(self._tax_now), "Type": "DubiousMergedfragment"})
                    list_member_first_fragments = list_member_first.split("_|_")
                    for fragment in list_member_first_fragments:
                        prot_name_integer = fragment.split("||")[2].strip()
                        geneRef_elemnt = ET.Element('geneRef', attrib={'id': str(prot_name_integer)})
                        paralog_element.append(geneRef_elemnt)
                    return paralog_element
                else:
                    prot_name_integer = list_member_first.split("||")[2].strip()
                    geneRef_elemnt = ET.Element('geneRef', attrib={'id': str(prot_name_integer)})
                    #'id': str(gene_id_name[list_member_first])})  # # gene_id_name[query_prot_record.id]   # hog_elemnt.append(geneRef_elemnt)
                    # to do could be improved when the rhog contains only one protein
                    return geneRef_elemnt

            elif len(list_member) > 1 and fragment_detection and (not fragment_detection_msa_merge):
                paralog_element = ET.Element('paralogGroup')
                property_element = ET.SubElement(paralog_element, "property", attrib={"name": "Type", "value":"Dubiousfragment"})
                # property_element = ET.SubElement(paralog_element, "property", attrib={"TaxRange": str(self._tax_now),"Type": "DubiousMergedfragment"})
                # removed proteins which are dubious are reported only in the log files.
                for member in list_member:
                    prot_name_integer = member.split("||")[2].strip()
                    geneRef_elemnt = ET.Element('geneRef', attrib={'id': str(prot_name_integer)})
                    paralog_element.append(geneRef_elemnt)
                return paralog_element

        # We continue this function as an Implicit else :  len(self._subhogs) >=1

        def _sorter_key(sh):
            return sh._tax_now.name
        self._subhogs.sort(key=_sorter_key)

        element_list = []
        for sub_clade, sub_hogs in itertools.groupby(self._subhogs, key=_sorter_key):  # sub_clade is the taxrange
            list_of_subhogs_of_same_clade = list(sub_hogs)
            # list_of_subhogs_of_same_clade is  [object of HOGclass hogID=HOG_D0634402_sub10094,len=12, tax_least=Amniota, tax_now= Amniota, object of HOGclass hogID=HOG_D0634402_sub10096,len=20, tax_least=Amniota, tax_now= Amniota, object of HOGclass hogID=HOG_D0634402_sub10081,len=1, tax_least=MONDO, tax_now= Amniota, object of HOGclass hogID=HOG_D0634402_sub10093,len=12, tax_least=Amniota, tax_now= Amniota, object of HOGclass hogID=HOG_D0634402_sub10095,len=7, tax_least=Amniota, tax_now= Amniota]
            # following only for debugging, can be deleted later
            for subhog in list_of_subhogs_of_same_clade:
                if len(subhog._members) == 0:
                    logger.warning("issue 12314506" + str(subhog) + str(sub_clade))

            if len(list_of_subhogs_of_same_clade) > 1:
                paralog_element = ET.Element('paralogGroup')
                # todo add score how much overlap they have, report as score
                # the following line could be improved, instead of tax_now we can use the least common ancestor of all members
                # property_element = ET.SubElement(paralog_element, "property",attrib={"name": "TaxRange", "value": str(sub_clade)}) # self._tax_now
                for sh in list_of_subhogs_of_same_clade:
                    element_p = sh.to_orthoxml()
                    if str(element_p):
                        paralog_element.append(element_p)  # ,**gene_id_name  indent+2
                    else:
                        logger.warning("issue 123434" + str(sh) + str(sub_clade))

                element_list.append(paralog_element)
            elif len(list_of_subhogs_of_same_clade) == 1:
                subhog = list_of_subhogs_of_same_clade[0]
                if len(subhog._members):
                    element = subhog.to_orthoxml()
                    if str(element):  # element could be  <Element 'geneRef' at 0x7f7f9bacb450>
                        element_list.append(element)  # indent+2
                    else:
                        logger.warning("issue 12359 "+str(subhog))
                else:
                    logger.warning("issue 12369 " + str(subhog))

        if len(element_list) == 1:
            return element_list[0]

        elif len(element_list) > 1:
            #hog_elemnt = ET.Element('orthologGroup', attrib={"id": str(self._hogid)})
            hog_elemnt = ET.Element('orthologGroup', attrib={"id": str(self._hogid)}, )
            num_species_tax_hog = len(set([i.split("||")[1] for i in self._members]))  #  'tr|H2MU14|H2MU14_ORYLA||ORYLA||1056022282'
            completeness_score = round(num_species_tax_hog/self._num_species_tax_speciestree,4)
            property_element = ET.SubElement(hog_elemnt, "score", attrib={"id": "CompletenessScore", "value": str(completeness_score)})
            property_element = ET.SubElement(hog_elemnt, "property", attrib={"name": "TaxRange", "value": str(self._tax_now.name)})

            for element in element_list:
                hog_elemnt.append(element)

            return hog_elemnt

""" 
        Example of orthoxml format
        <orthologGroup id="HOG_0022401_sub10169">
           <property name="TaxRange" value="Alcidae"/>
           <paralogGroup>
              <geneRef id="1163015305"/>
              <orthologGroup id="HOG_0022401_sub10166">
                 <property name="TaxRange" value="Alca_torda-Uria"/>
                 <orthologGroup id="HOG_0022401_sub10164">
                    <property name="TaxRange" value="Uria"/>
                    <geneRef id="1164015054"/>
                    <geneRef id="1180013314"/>
                 </orthologGroup>
              </orthologGroup>
           </paralogGroup>
        </orthologGroup>
"""


def split_hog(hog:HOG, *partitions):
    """splits a hog into parts based to the partitions provided.

    The partitions need to be a lists of Representatives (or simply members)
    of the given HOG.

    :param partitions:  a list or a tuple of Representatives of the current HOG
    :type partitions:   List[Representatives]
    :returns List[HOG]: a list of HOG objects splitting the original HOG in the
                        desired partitions.
    """
    max_depth = 4
    if len(partitions) < 2:
        raise ValueError(f"Must provide at least two partitions to split {hog}")
    partitions = [set(hog.find_representative(r) for r in part) for part in partitions]
    if not all(p1.isdisjoint(p2) for p1, p2 in itertools.combinations(partitions, 2)):
        raise ValueError(f"Not all partitions are non-overlapping {partitions}")
    #if sum(len(p) for p in partitions) != len(set(hog.get_representatives())):
    #    raise ValueError(f"Partitions to split {hog} must cover all {len(hog.get_representatives())} representatives, but cover only {sum(len(p) for p in partitions)}.")
    rep_to_subhog_list = {rep: hog.get_subhog_path(rep.get_id(), max_depth) for part in partitions for rep in part}

    logger.debug(f"Subhog paths of represenatatives for {hog}")
    for i, part in enumerate(partitions):
        logger.debug(f"---Partition {i} -----")
        for rep in part:
            path = " --> ".join(str(h) for h in rep_to_subhog_list[rep])
            logger.debug(f"{rep.get_id()}: {path}")

    first_level_subhogs = [set(rep_to_subhog_list[rep][0] for rep in part) for part in partitions]
    if all(s1.isdisjoint(s2) for s1, s2 in itertools.combinations(first_level_subhogs, 2)):
        logger.info(f"Splitting {hog} into {len(first_level_subhogs)} subhogs: {first_level_subhogs}")
        hogs = []
        for p, subhogs in enumerate(first_level_subhogs):
            h = HOG(subhogs, taxnomic_range=hog.taxlevel, rhogid=hog.rhogid, msa=hog.get_msa(), representatives=partitions[p])
            hogs.append(h)
        return hogs
    #raise RuntimeError("this part of the code needs more thinking")



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.builtins import dict
from future.builtins import zip
from future.builtins import range
from future import standard_library
standard_library.install_hooks()

try:
    from progressbar import ProgressBar, Percentage, Timer, ETA, Bar
    PROGRESSBAR = True
except ImportError:
    PROGRESSBAR = False

from collections import deque

def setup_progressbar(msg, size):
    if not msg.endswith(': '):
        msg += ': '

    widgets = [msg,
               Percentage(), ' ',
               Bar(), ' ',
               Timer(), ' ',
               ETA()]

    pbar = ProgressBar(widgets=widgets, maxval=size)
    return pbar

def enum(*sequential, **named):
    """creates an Enum type with given values"""
    enums = dict(zip(sequential, range(len(sequential))), **named)
    enums['reverse'] = dict((value, key) for key, value in enums.items())
    return type('Enum', (object, ), enums)


class IterableClassException(Exception):
    pass

def py2_iterable(Class):
    """
    Use as a class decorator to make a class that has a python 3 next method --
    __next__() -- also iterable with python 2, which uses next(). Also checks
    for an __iter__ method -- if this is missing the class won't be iterable anyway.


    e.g.
    @py2_iterable
    class Py2and3Iterator(object):
        def __init__(self):
            self.data = list('somestuff')
            self._pos = 0

        def __iter__(self):
            return self

        def __next__(self):
            if self._pos == len(self.data):
                self._pos = 0
                raise StopIteration
            char = self.data[self._pos]
            self._pos += 1
            return char


    :param Class: the class being decorated
    :return: Class: the decorated class, which is iterable in py2 and py3
    """
    if not hasattr(Class, '__iter__'):
        raise IterableClassException('Class "{}" has no __iter__ method and will not be iterable'
                                     .format(Class.__class__.__name__))

    if hasattr(Class, '__next__'):
        next_method = getattr(Class, '__next__')
        setattr(Class, 'next', next_method)

    return Class


@py2_iterable
class Queue(object):

    def __init__(self):
        self.__queue = deque()

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.__queue)

    def __next__(self):
        if self.isempty():
            raise StopIteration
        return self.dequeue()

    def enqueue(self, item):
        self.__queue.append(item)

    def dequeue(self):
        if self.isempty():
            raise Exception('empty queue')
        return self.__queue.popleft()

    def isempty(self):
        return len(self.__queue) == 0
